# -*- coding: utf-8 -*-

#  PycaWM
#  remoteREPL.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# 'with' stuff
from __future__ import with_statement

import sys

if sys.version_info < (2, 5):
    raise ImportError('remoteREPL module needs at least Python 2.5!')

import __builtin__
import __main__
import os
import socket
import struct
import threading

from code import InteractiveConsole
from contextlib import contextmanager
from cPickle import dumps as pickle_dumps
from keyword import kwlist
from re import match as re_match

from pycawm.wrappers import SocketWrapper

SOCK_NAME = '/tmp/pycawm-%s' % os.getuid()

class RemoteREPL(InteractiveConsole):
    def __init__(self, wm):
        InteractiveConsole.__init__(self)
        self.wm = wm
        self.server_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.server_socket.bind(SOCK_NAME)
        self.server_socket.listen(5)
        self.socket_wrapper = SocketWrapper(self.server_socket,
                                            self.handle_connection)
        self.sockets = wm.sockets
        self.sockets.append(self.socket_wrapper)

    def destroy(self):
        self.server_socket.close()
        try:
            self.sockets.remove(self.socket_wrapper)
        except ValueError:
            pass

    def remote_interact(self, sock, banner):
        self.locals = __main__.__dict__
        self.sock = sock
        self.write_rlock = threading.RLock()
        try:
            self.interact(banner)
        except:
            self.sock.close()
            self.sockets.append(self.socket_wrapper)

    @contextmanager
    def stdout_networked(self):
        sys.stdout = self
        sys.stderr = self
        try:
            yield None
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__

    def push(self, line):
        with self.write_rlock:
            with self.stdout_networked():
                return InteractiveConsole.push(self, line)

    def global_matches(self, text):
        matches = list()
        for list_ in [kwlist,
                      __builtin__.__dict__.keys(),
                      __main__.__dict__.keys()]:
            for word in list_:
                if word.startswith(text) and word != '__builtins__':
                    matches.append(word)
        matches_str = pickle_dumps(matches)
        self.write(matches_str, 'G')

    def attr_matches(self, text):
        matches = list()
        m = re_match(r'(\w+(\.\w+)*)\.(\w*)', text)
        if m:
            expr, attr = m.group(1, 3)
            try:
                obj = eval(expr, __main__.__dict__)
                words = dir(obj)
                if hasattr(obj, '__class__'):
                    words.append('__class__')
                    words = words + get_class_members(obj.__class__)
                for word in words:
                    if word.startswith(attr) and word != '__builtins__':
                        matches.append('%s.%s' % (expr, word))
            except NameError:
                pass
        matches_str = pickle_dumps(matches)
        self.write(matches_str, 'A')

    def runcode(self, code):
        with self.wm.event_lock:
            InteractiveConsole.runcode(self, code)
            self.wm.display.sync()

    def interact(self, banner):
        if not hasattr(sys, 'ps1'):
            sys.ps1 = '>>> '
        if not hasattr(sys, 'ps2'):
            sys.ps2 = '... '
        self.write('%s\n' % str(banner))
        self.write(sys.ps1, 'P')
        while 1:
            try:
                try:
                    line = self.raw_input()
                except EOFError:
                    self.write('\n')
                    break
                else:
                    if line is not None:
                        more = self.push(line)
                        if more:
                            prompt = sys.ps2
                        else:
                            prompt = sys.ps1
                        self.write(prompt, 'P')
            except KeyboardInterrupt:
                self.write('\nKeyboardInterrupt\n')
                self.resetbuffer()

    def recv_data(self):
        # retrieve size of the data
        size = struct.calcsize('I')
        size = self.sock.recv(size)
        # retrieve data
        size = struct.unpack('<I', size)[0]
        data = self.sock.recv(size)
        return data

    def send_data(self, text, type_):
        data_sent = type_ + text
        size = struct.pack('<I', len(data_sent))
        self.sock.sendall(size + data_sent)

    def raw_input(self):
        data = self.recv_data()
        if data[0] == 'R':
            return data[1:]
        elif data[0] == 'G':
            self.global_matches(data[1:])
            return None
        elif data[0] == 'A':
            self.attr_matches(data[1:])
            return None

    def write(self, data, type_='D'):
        with self.write_rlock:
            self.send_data(data, type_)

    def handle_connection(self):
        self.sockets.remove(self.socket_wrapper)
        repl = threading.Thread(target=self.remote_interact,
                                args=(self.server_socket.accept()[0],
                                      'Welcome to the PycaWM REPL!'))
        repl.setDaemon(True)
        repl.start()

def get_class_members(klass):
    ret = dir(klass)
    if hasattr(klass, '__bases__'):
        for base in klass.__bases__:
            ret = ret + get_class_members(base)
    return ret
